/**
 * Copyright (C) Mellanox Technologies Ltd. 2018.       ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

/* *******************************************************
 * LOONGARCH64 processors family                                 *
 * ***************************************************** */
#if defined(__loongarch__)

#include <sys/mman.h>
#include <dlfcn.h>
#include <string.h>
#include <stdlib.h>

#include <ucm/bistro/bistro.h>
#include <ucm/bistro/bistro_int.h>
#include <ucm/util/sys.h>
#include <ucs/sys/math.h>
#include <ucs/arch/cpu.h>
#include <ucs/debug/assert.h>


/* Register number used to store indirect jump address.
 * r20 is the highest numbered temporary register, assuming this one is safe
 * to use. */
#define R20 20

/**
 * @brief Generate a mov immediate instruction
 *
 * @param[in] _rd, _rj  register number (0-31)
 * @param[in] _si20, _si20 signed mmediate value 20, 12 bits
 * @param[in] _ui12 immediate value
 */
#define LU12I_W(_rd, _si20) \
    ((10 << 25) + (((_si20) & 0xfffff) << 5) + (_rd))

#define LU32I_D(_rd, _si20) \
    (((22 >> 1) << 25) + (((_si20) & 0xfffff) << 5) + (_rd))

#define LU52I_D(_rd, _rj, _si12) \
    (((48 >> 2) << 22) + (((_si12) & 0xfff) << 10) + ((_rj) << 5) + (_rd))

#define ORI(_rd, _rj, _ui12) \
    (((56 >> 2) << 22) + (((_ui12) & 0xfff) << 10) + ((_rj) << 5) + (_rd))

#define JIRL(_rd, _rj, _offs) \
    ((19 << 26) + ((_offs) << 10) + ((_rj) << 5) + (_rd))

ucs_status_t ucm_bistro_patch(void *func_ptr, void *hook, const char *symbol,
                              void **orig_func_p,
                              ucm_bistro_restore_point_t **rp)
{
    ucm_bistro_patch_t patch = {
        .lu12i = LU12I_W(R20, (uintptr_t)hook >> 12),
        .ori = ORI(R20, R20, (uintptr_t)hook),
        .lu32i = LU32I_D(R20, (uintptr_t)hook >> 32),
        .lu52i = LU52I_D(R20, R20, (uintptr_t)hook >> 52),
        .jirl = JIRL(0, R20, 0)
    };
    ucs_status_t status;

    if (orig_func_p != NULL) {
        return UCS_ERR_UNSUPPORTED;
    }

    status = ucm_bistro_create_restore_point(func_ptr, sizeof(patch), rp);
    if (UCS_STATUS_IS_ERR(status)) {
        return status;
    }

    return ucm_bistro_apply_patch(func_ptr, &patch, sizeof(patch));
}

ucs_status_t ucm_bistro_relocate_one(ucm_bistro_relocate_context_t *ctx)
{
    return UCS_ERR_UNSUPPORTED;
}
#endif
